/*
                                bj.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*       Author:         E.BERTIN        V1.0    (04/09/93)
*
*       Purpose:        use Aoki's matrix for B1950->j2000..
*
*       Requirements:   to be linked with the math library (-lm).
*
*       Calculations:   variable.
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define         PI              3.141592653589793
#define         ARCSEC          PI/(180.0*3600.0)
#define         SYNTAX  \
"b2j -o <year_obs> [options] <filein >fileout\n\n\
 options are:           -q (quiet flag: defaulted to no)\n"

#define         QPRINTF(x)      if (!qflag)     fprintf(stderr,x)


/*------------------------------- structures
--------------------------------*/

/*----------------------- miscellaneous variables
---------------------------*/
char            qflag;

/*------------------------------- functions
---------------------------------*/
void            bj(), error(), precess();

/********************************* bj
***********************************/
/*
conver equatorial coordinates from equinox and epoch B1950 to equinox
and
epoch J2000 for extragalactic sources (from Aoki et al. 1983).
*/
void    bj(yearobs, alphain, deltain, alphaout, deltaout)
double  yearobs, alphain, deltain, *alphaout, *deltaout;
        {
        int             i,j;
        static double   a[3] = {-1.62557e-6, -0.31919e-6, -0.13843e-6},
                        ap[3] = {1.245e-3, -1.580e-3, -0.659e-3},
                        m[6][6] = {     0.9999256782,   
-0.0111820611,    -0.0048579477,
0.00000242395018,-0.00000002710663,-0.00000001177656,
0.0111820610,    0.9999374784,    -0.0000271765,0.00000002710663,
0.00000242397878,-0.00000000006587,
0.0048579479,   -0.0000271474,     0.9999881997,
0.00000001177656,-0.00000000006582, 0.00000242410173,
-0.000551,       -0.238565,         0.435739,
0.99994704,     -0.01118251,      -0.00485767,
0.238514,       -0.002662,        -0.008541,
0.01118251,      0.99995883,      -0.00002718,
0.012254,          0.002117,     0.00485767,    
-0.00002714,        1.0000956};
        double          a1[3], r[3], ro[3], r1[3], r2[3], v1[3], v[3],
                        cai, sai, cdi, sdi, dotp, rmod, alpha, delta,
                        t1 = (yearobs - 1950.0)/100.0;

        alphain *= PI/180.0;
        deltain *= PI/180.0;
        cai = cos(alphain);
        sai = sin(alphain);
        cdi = cos(deltain);
        sdi = sin(deltain);
        ro[0] = cdi*cai;
        ro[1] = cdi*sai;
        ro[2] = sdi;
        dotp = 0.0;
        for (i=0; i<3; i++)
                {
                a1[i] = a[i]+ap[i]*ARCSEC*t1;
                dotp += a1[i]*ro[i];
                }
        for (i=0; i<3; i++)
                {
                r1[i] = ro[i] - a1[i] + dotp*ro[i];
                r[i] = v[i] = v1[i] = 0.0;
                }
        for (j=0; j<6; j++)
                for (i=0; i<6; i++)
                        {
                        if (j<3)
                                r[j] += m[j][i]*(i<3?r1[i]:v1[i-3]);
                        else
                                v[j-3] += m[j][i]*(i<3?r1[i]:v1[i-3]);
                        }
        rmod = 0.0;
        for (i=0; i<3; i++)
                {
                r2[i] = r[i]+v[i]*ARCSEC*(t1-0.5);
                rmod += r2[i]*r2[i];
                }
        rmod = sqrt(rmod);
        delta = asin(r2[2]/rmod);
        alpha = acos(r2[0]/cos(delta)/rmod);
        if (r2[1]<0)
                alpha = 2*PI - alpha;
        *alphaout = alpha*180.0/PI;
        *deltaout = delta*180.0/PI;

        return;
        }

/********************************* precess
***********************************/
/*
precess equatorial coordinates according to the equinox (from
Ephemerides du
Bureau des Longitudes 1992). Epoch for coordinates should be J2000
(FK5 system).
*/
void    precess(yearin, alphain, deltain, yearout, alphaout, deltaout)
double  yearin, alphain, deltain, yearout, *alphaout, *deltaout;

        {
        double  dzeta,theta,z, t1,t2, cddsadz, cddcadz, cdd, sdd, adz;

        alphain *= PI/180.0;
        deltain *= PI/180.0;

        t1 = (yearin - 2000.0)/1000.0;
        t2 = (yearout - yearin)/1000.0;
        theta = (97171.735e-06 - 413.691e-06*t1 - 1.052e-06 * t1*t1) *
t2
            + (-206.846e-06 - 1.052e-06*t1) * t2*t2 - 202.812e-06 *
t2*t2*t2;
        dzeta = (111808.609e-06 + 677.071e-06*t1 - 0.674e-06 * t1*t1) *
t2
            + (146.356e-06 - 1.673e-06*t1) * t2*t2 + 87.257e-06 *
t2*t2*t2;
        z = (111808.609e-06 +677.071e-06*t1 - 0.674e-06 * t1*t1) * t2
            + (530.716e-06 + 0.320e-06*t1) * t2*t2 + 88.251e-06 *
t2*t2*t2;
        cddsadz = cos(deltain) * sin(alphain+dzeta);
        cddcadz = -sin(deltain)*sin(theta)
                + cos(deltain)*cos(theta)*cos(alphain+dzeta);
        sdd = sin(deltain)*cos(theta)
                + cos(deltain)*sin(theta)*cos(alphain+dzeta);
        cdd = cos(*deltaout = asin(sdd));
        adz = asin(cddsadz/cdd);
        if (cddcadz<0.0)
                adz = PI - adz;
        if (adz<0.0)
                adz += 2.0*PI;
        adz += z;
        *alphaout = 180.0/PI*adz;
        *deltaout *= 180.0/PI;
        }

/********************************* error
************************************/
/*
I hope the (pc) counter won't reach that point!
*/
void    error(num,msg1,msg2)
short   num;
char    *msg1, *msg2;
        {
        fprintf(stderr,"\n> %s%s\n\n",msg1,msg2);
        exit(num);
        }

/********************************** main
************************************/

main(argc, argv)
int     argc;
char    *argv[];
        {
        char    s[256];
        double  alpha, delta, yearobs;
        int     a;

/* default parameters */
        qflag = 0;
        yearobs = 1950.0;

        if (argc>1)
          for (a=1; a<argc; a++)
           switch(tolower(argv[a][1]))
                {
                case 'o':       yearobs = atof(argv[++a]);
                                break;
                case 'q':       qflag = 1;
                                break;
                default:        error(0,"SYNTAX: ", SYNTAX);
                                break;
                }

        QPRINTF ("\n\n                            BJ\n\n");
        QPRINTF ("\nconvert B1950 into J2000 coordinates\n");
        QPRINTF ("\nE. BERTIN 1993\n\n");

        while(gets(s))
                {
                sscanf(s, "%lf %lf", &alpha, &delta);
                bj(yearobs, alpha, delta, &alpha, &delta);
                printf("%10.6lf %10.6lf\n", alpha,delta);
                }
        return;
        }
