/*++++++++++++++
.IDENTIFICATION aclient.c
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [CDS]
.ENVIRONMENT    
.KEYWORDS       
.VERSION  1.0   23-Jul-1992
.VERSION  1.1   12-Nov-1992: Added parameters -b -p
.VERSION  2.0   10-May-1993: File Redirection, parameters -U -P
.VERSION  2.1   18-Aug-1993: Possible Interruption.
.VERSION  2.2   25-Sep-1993: Allow pipes as intput/output
.VERSION  2.3   06-Feb-2003: Accept commands from stdin 
.VERSION  2.4   12-avr-2003: Add a timeout (default 60sec) for the connection.
.VERSION  2.5   09-Jun-2004: Use the getpass function by default.
.VERSION  2.6   10-Mar-2006: -version option
.COMMENTS       Dialogue with a server, using the ^D / ^B / ^F Conventions.
		(see "sk" man pages)
---------------*/

#define VERSION "2.6"

#include <sk.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>	/* INTR is sent to Server */
#include <setjmp.h>	/* To exit from Interrupt */
#include <stdlib.h>	/* malloc */
#include <sys/types.h>	/* malloc */
#include <sys/stat.h>	/* malloc */
#ifdef USE_TERMIO
#include <termio.h>	/* The following is for TermIO */
#ifdef TCGETS
# define termio termios
# define TCget	TCGETS
# define TCset	TCSETSW
#else
# define TCget	TCGETA
# define TCset	TCSETAW
#endif
#endif

#ifndef TIME_MAX
#define TIME_MAX      60	/* max. time to connect */
#endif

#define MIN(a,b)	((a) < (b) ? (a) : (b))
#define ITEMS(a)	(sizeof(a)/sizeof(a[0]))

static char usage[] = 
 "Usage: %s [-b blksiz] [-bsk sksiz] [-p prompt] host service -Uname\
 -Ppasswd\n";
static char pgm[]  = "aclient";
static int fdigest = 1;		/* File number where to write */
static int fmore   = 0;		/* File number where to get   */
static int blksize = 4096;	/* Standard blocksize	      */
static char *username;
static char *password;
static jmp_buf jmp_env;
static int interrupted;
static char *connect_err[] = {
    "can't connect", 		/* -1 */
    "connection refused"	/* -2 */
};


/*===========================================================================*/
static char *getpasswd()
/*++++++++++++++++
.PURPOSE  Get a pawwsord at Terminal
.RETURNS  Number of characters
-----------------*/
{
#ifdef USE_TERMIO
  static char buf[20];	/* Buffer which will contain Password, limited */
  static char pass[] = "Password: ";
  struct termio tty0, tty1;
  int n;

    ioctl(0, TCget, &tty0); 
    ioctl(0, TCget, &tty1); 
    tty1.c_lflag &= ~ECHO; 
    ioctl(0, TCset, &tty1); 	/* Turn ECHO off */
    write(1, pass, strlen(pass));
    n = read(0, buf, sizeof(buf));
    ioctl(0, TCset, &tty0); 	/* Reset to previous */
    if (n <= 0) perror("****Get Password"), n = 1;
    buf[--n] = 0;
    return(buf);
#else
  char *getpass();
    return(getpass("Password: "));
#endif
}

/*===========================================================================*/
static void help()
{
  static char *text[] = {
	" -b blksiz: size of blocks to send to Server",
	" -bsk blks: size of socket blocks (default $SK_bsk)",
	" -p prompt: prompt to be used in interactive mode", 
	"      host: Name of Internet Number of service's host",
	"   service: service name or number",
	"    -Uname: Username for the Server",
	"    -Ppass: Password for the Server",
  };
  int i;

  printf(usage, pgm);
  for (i=0; i < ITEMS(text); i++)	
	printf("%s\n", text[i]);
  exit(0);
}

static int digest(buf, len)
/*++++++++++++++++
.PURPOSE  Function to write to fdigest
.RETURNS  Number of bytes written
-----------------*/
    char *buf;	/* IN: Buffer with data */
    int  len;	/* IN: How many bytes   */
{
	int n;

  if (interrupted) return(0);		/* don't write... */
  if (len>0) n = write(fdigest, buf, len);
  else n = 0;
  if (n < 0) perror("****aClient Digest");
  return(n);
}

static int more(buf, len)
/*++++++++++++++++
.PURPOSE  Function to read blocks
.RETURNS  Number of bytes written
-----------------*/
    char *buf;	/* IN: Buffer to write   */
    int  len;	/* IN: Bytes written out */
{
	int n;
	char *p, *pe;

  if (interrupted) return(0);		/* Emulate End of File */
  pe = buf + MIN(blksize, len);
  for (p = buf; p < pe; p += n) {
  	n = read(fmore, p, pe-p);
  	if (n < 0) perror("****aClient / More");
	if (n <= 0) break;
  }
  return(p-buf);
}

static int myopen(fname, mode, scanned)
/*++++++++++++++++
.PURPOSE  Open a file with error report
.RETURNS  File Handle. The length used is stored on output.
-----------------*/
  char *fname;	/* IN: File name (only first word)	*/
  int  mode;	/* IN: Mode (Read/Write/Append)		*/
  int *scanned;	/* OUT: How many bytes have been scanned*/
{
  char *p, *pname, x, quote; int fh;
  static char *popmode[2] = {"r", "w"}; FILE *popfile;

    pname = fname;
    quote = *pname;
    if ((quote == '\'') || (quote == '"')) {	/* Quoted filename */
    	for (p = ++pname; *p && (*p != quote); p++) ;
    }
    else {
	quote = ' ';
    	for (p=pname; isgraph(*p); p++) ;
    }
    if (*pname == '|') { /* We've to open a PIPE */
	popfile = (FILE *)1;
	while(*++pname == ' ');
    }
    else popfile = (FILE *)0;
    x = *p; *p = 0; 

    if (popfile) {
	popfile = popen(pname, popmode[mode&1]);
	if (popfile) fh = fileno(popfile);
	else fh = -1;
    }
    else fh = open(pname, mode);
    *p = x; 
    *scanned = p-fname;
    if (x == quote) *scanned += 1;
    if (fh < 0) { perror(fname); exit(1); }
    return(fh);
}

static int theplug;	/* For communication by signals */
/*===========================================================================*/
static void onIntr(s)
/*++++++++++++++++
.PURPOSE  Activated when INTR
.RETURNS  --
-----------------*/
  int s;	/* IN: Signal number */
{
#ifdef DEBUG
    printf("+++ onIntr called, signal %d, theplug=%d\n", s, theplug);
#endif
    signal(s, onIntr);
    sk_kill(theplug, 0);
    interrupted = 1;
    /* longjmp(jmp_env, 1); */
}

static void tooLong(s)
/*++++++++++++++++
.PURPOSE  Activated when timeout away
.RETURNS  --
-----------------*/
  int s;	/* IN: Signal number */
{
  static int ncalls = 0;
    ncalls++;
#ifdef DEBUG
    printf("+++ tooLong#%d called, signal %d, theplug=%d\n", 
      ncalls, s, theplug);
#endif
    fprintf(stderr, "****aclient: too long to connect (TIME_MAX=%d)\n",
      TIME_MAX);
    exit(1);
}

/*===========================================================================*/
int main(argc,argv)
int argc; char *argv[];
{

    int plug;		/* socket to "plug" into the socket */
    char buf[BUFSIZ], exec_0, *p, *host, *service, *parm, *prompt;
    int i, tty, blk, fop, got_command, scanned;
    char bs;
    char *getenv();

    prompt = host = service = (char *)0;
    /* sk_setlog(NULL); */

	/* Define umask for file creation */
    umask(022);

	/* SK_bsk env. variable may tell the Physical Blocksize */

    if ((p = getenv("SK_bsk"))) {
	i = atoi(p);
	while(isdigit(*p)) p++;
	if (tolower(*p) == 'K') i *= 2;
	sk_setb(i);
    }
	/* Examine Arguments */

    while (-- argc > 0) {
	parm = *++argv;
	/* fprintf(stderr, "... Examine %s\n", parm); */
	if (*parm == '-') switch (*++parm) {
	  case 'b':	/* Blocksize */
	    bs = parm[1];	/* bs = 's' for Socket Blksize */
	    parm = *++argv; --argc;
	    for (p = parm; isdigit(*p); p++) ;	/* Terminates by k or b */
	    blk = atoi(parm);
	    switch(*p) {
	      case 'k': case 'K':
		blk *= 1024; break;
	      case 'b': case 'B':
		blk *= 512; break;
	      case '\0':	break;
	      default:
		blk = 0;    break;
	    }
	    if (blk <= 0) {
		printf("*** %s: Bad b (blocksize) parameter: %s\n",pgm, parm);
		exit(__LINE__);
	    }
	    if (bs == 's')	sk_setb(blk);
	    else  blksize = blk;
	    continue;
	  case 'p':	/* Prompt    */
	    prompt = *++argv; --argc;
	    continue;
	  case 'h':
	    help();
	    exit(0);
          case 'U':	/* UserName */
            username = parm+1;
            continue;
          case 'P':	/* Password */
            password = parm+1;
            *parm = 0; 	/* Hide it  */
            continue;
	  case 'v':
	    if (strncmp(parm, "ver", 3) == 0) {
		printf("aclient (CDS) Version %s\n", VERSION);
		exit(0);
	    }
	    goto bad_option;
	  default:
	    if (host && service) break;
	  bad_option:
	    printf("****%s: which option is -%s ? ", pgm, parm);
	    help();
	    exit(__LINE__);
	}
	if (!host)    host = parm;
	else if (!service) service = parm; 
	else { --argv; ++argc; break;}	/* We've now commands */
    }

    /* Be sure the connection does not wait for ever ! */
    signal(SIGALRM, tooLong); alarm(TIME_MAX);

    /* Check the information */
    if ((!host) || (!service)) {
        printf(usage, pgm);
        exit(__LINE__);
    }
    if (!prompt) {
	prompt = malloc(strlen(host)+strlen(service)+3);
	strcpy(prompt, host); strcat(prompt, "/"); strcat(prompt, service);
    }

    if (username) {	/* Prompt for a Password if necessary */
	if (!password && isatty(0)) password = getpasswd();
	plug = sk_connect(host, service, username, password);
    }
    else plug = sk_open(host, service);
    if (plug < 0 ) {
	printf("   (Returned code = %d", plug);
	plug = -1-plug;
	if (plug < ITEMS(connect_err)) printf(": %s", connect_err[plug]);
	printf(")\n");
	exit(__LINE__);
    }
    theplug = plug;
    alarm(0);		/* Stop the alarm */

	/* Take command from argument line */
    exec_0 = argc <= 1;		/* Indicator to get commands from stdin */
    tty = isatty(0);
 
	/* If no command found in Argument line, get one now */
  GetCommand: 
    signal(SIGINT, onIntr);
    setjmp(jmp_env);
    interrupted = 0;
    buf[0] = '\04';		/* END indicator */
    if (exec_0) {
    	if(tty) printf("%s> ", prompt);
    	if (!fgets(buf, sizeof(buf), stdin)) buf[0] = '\04';
	else for (p=buf+strlen(buf)-1; (p>=buf)&&iscntrl(*p); p--) *p = 0;
    }
    else for (p = buf; --argc > 0; p += strlen(p)) {
	parm = *++argv;
	if (strcmp(parm, ";") == 0) { *p = 0; break; }
	if (p != buf) *(p++) = ' ';
	strcpy(p, parm);
    }
    p = buf; fdigest = 1; fmore = 0;
    for (got_command = 0; ! got_command; ) {
    	while (isspace(*p)) p++;
    	switch(*p) {
    	  case '\04': exit(0);
    	  case '<':		/* Input from a file */
    	    p++; while (isspace(*p)) p++;
    	    fmore = myopen(p, 0, &scanned);
    	    p += scanned;
    	    continue;
          case '>':		/* Output Redirection */
            fop = 1;			/* Write  */
            if (*++p == '>') { 
            	fop = 2;		/* Append */
            	p++;
	    }
	    while(isspace(*p)) p++;
	    fdigest = myopen(p, fop|O_CREAT, &scanned);
    	    p += scanned;
    	    continue;
          case '#':			/* Comment */
            puts(p); 
            goto GetCommand;
    	  default: got_command = 1;
    	}
    }

#ifdef DEBUG
    printf(".... sk_puts(%d, <%s>)\n", plug, p);
#endif
    if ( sk_puts(plug, p) < 0 ) exit(__LINE__);
    /* fprintf(stderr, ".... Execute <%s>\n", buf); */

    if (sk_obeyserver(plug, digest, more) < 0) exit(1);
    if (fdigest > 2) close(fdigest);
    if (fmore   > 2) close(fmore  );

	/* Loop to next command */
    goto GetCommand;
}
